/*
  * Copyright (c) 2019 Ignacio Francisco Ramírez Paulino and Ignacio Hounie
  * This program is free software: you can redistribute it and/or modify it
  * under the terms of the GNU Affero General Public License as published by
  * the Free Software Foundation, either version 3 of the License, or (at
  * your option) any later version.
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
  * General Public License for more details.
  * You should have received a copy of the GNU Affero General Public License
  *  along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
/**
 * \file paco_linop.c
 * \brief linear operators defined on patches
 */
#include <omp.h>
#include <assert.h>
#include <gsl/gsl_matrix_float.h>
#include <gsl/gsl_linalg.h>
#include <fftw3.h>

#include "paco_types.h"
#include "paco_log.h"

//----------------------------------------------------------------------------

#define MAX_SIZE
#define MAX_THREADS 64 /// if I ever get to reach this limit I'll be very happy

#ifdef _OPENMP
//
// OMP available: wrap functions
//
static int get_num_threads() {
    int NT = 1;
    #pragma omp parallel
    {
        NT = MAX_THREADS > omp_get_num_threads() ? omp_get_num_threads(): MAX_THREADS;
        omp_set_num_threads(NT);
    }
    return NT;
}

static int get_thread_num() {
    return omp_get_thread_num();
}

#else
//
// no OMP: provide simple stubs
//
static int get_num_threads() {
    return 1;
}
static int get_thread_num() {
    return 0;
}

#endif


//---------------------------------------------------------------------------------------

void dct2d(gsl_matrix_float* TX, const index_t w1, const index_t w2, const gsl_matrix_float* X) {
    sample_t* bufin[MAX_THREADS]; // first, hopefully aligned!
    sample_t* bufout[MAX_THREADS]; // first, hopefully aligned!
    fftwf_plan plan[MAX_THREADS];
    const index_t m = TX->size2;
    const index_t n = TX->size1;
    assert(m == w1*w2);
    index_t NT = get_num_threads();

    for (unsigned T = 0; T < NT; T++) {
        bufin[T] = (sample_t*) malloc(m*sizeof(sample_t));
        bufout[T] = (sample_t*) malloc(m*sizeof(sample_t));
        plan[T] = fftwf_plan_r2r_2d(w1,w2,bufin[T],bufout[T],FFTW_REDFT10,FFTW_REDFT10,FFTW_ESTIMATE);
    }

#ifdef _OPENMP
    #pragma omp parallel for
#endif
    for (index_t j = 0; j < n; j++) {
        const index_t T = get_thread_num();
        fftwf_plan planT = plan[T];
        sample_t* inT = bufin[T];
        sample_t* outT = bufout[T];
        gsl_vector_float_view Xj = gsl_vector_float_view_array(inT,m);
        gsl_vector_float_view TXj = gsl_vector_float_view_array(outT,m);
        gsl_matrix_float_get_row(&Xj.vector,X,j);
        fftwf_execute_r2r(planT,inT,outT);
        gsl_matrix_float_set_row(TX,j,&TXj.vector);
    }

    for (unsigned T = 0; T < NT; T++) {
        fftwf_destroy_plan(plan[T]);
        free(bufout[T]);
        free(bufin[T]);
    }
    fftwf_cleanup();
}

//---------------------------------------------------------------------------------------

void idct2d(gsl_matrix_float* X, const index_t w1, const index_t w2, const gsl_matrix_float* TX) {
    sample_t* bufin[MAX_THREADS]; // first, hopefully aligned!
    sample_t* bufout[MAX_THREADS]; // first, hopefully aligned!
    fftwf_plan plan[MAX_THREADS];
    const index_t m = TX->size2;
    const index_t n = TX->size1;
    assert(m == w1*w2);
    const index_t NT = get_num_threads();

    for (unsigned T = 0; T < NT; T++) {
        bufin[T] = (sample_t*) malloc(m*sizeof(sample_t));
        bufout[T] = (sample_t*) malloc(m*sizeof(sample_t));
        plan[T] = fftwf_plan_r2r_2d(w1,w2,bufin[T],bufout[T],FFTW_REDFT01,FFTW_REDFT01,FFTW_ESTIMATE);
    }

#ifdef _OPENMP
    #pragma omp parallel for
#endif
    for (index_t j = 0; j < n; j++) {
        const index_t T = get_thread_num();
        fftwf_plan planT = plan[T];
        sample_t* inT = bufin[T];
        sample_t* outT = bufout[T];
        gsl_vector_float_view TXj = gsl_vector_float_view_array(inT,m);
        gsl_vector_float_view Xj = gsl_vector_float_view_array(outT,m);
        gsl_matrix_float_get_row(&TXj.vector,TX,j);
        fftwf_execute_r2r(planT,inT,outT);
        gsl_matrix_float_set_row(X,j,&Xj.vector);
    }

    for (unsigned T = 0; T < NT; T++) {
        fftwf_destroy_plan(plan[T]);
        free(bufout[T]);
        free(bufin[T]);
    }
    fftwf_cleanup();
}

//---------------------------------------------------------------------------------------

static void dct2n_prenormalize(sample_t* inT, index_t w1, index_t w2) {
}

//---------------------------------------------------------------------------------------

static void dct2n_postnormalize(sample_t* outT, index_t w1, index_t w2) {
    const index_t m = w1*w2;
    const sample_t k = 1.0/sqrt(2*w1*2*w2);
    const sample_t k00 = 0.5;
    const sample_t k01 = (sample_t) M_SQRT1_2;
    const sample_t k10 = (sample_t) M_SQRT1_2;
    // upper-left element

    for (index_t i = 0; i < m; ++i) // all
        outT[i] *= k;
    outT[0] *= k00;
    for (index_t j = 1; j < w2; ++j) // first row
        outT[j] *= k01;
    for (index_t i = 1; i < w1; ++i) // first column
        outT[w2*i] *= k10;
}

//---------------------------------------------------------------------------------------

void dct2dn(gsl_matrix_float* TX, const index_t w1, const index_t w2, const gsl_matrix_float* X) {
    sample_t* bufin[MAX_THREADS]; // first, hopefully aligned!
    sample_t* bufout[MAX_THREADS]; // first, hopefully aligned!
    fftwf_plan plan[MAX_THREADS];
    const index_t m = TX->size2;
    const index_t n = TX->size1;
    assert(m == w1*w2);
    index_t NT = get_num_threads();

    for (unsigned T = 0; T < NT; T++) {
        bufin[T] = (sample_t*) malloc(m*sizeof(sample_t));
        bufout[T] = (sample_t*) malloc(m*sizeof(sample_t));
        plan[T] = fftwf_plan_r2r_2d(w1,w2,bufin[T],bufout[T],FFTW_REDFT10,FFTW_REDFT10,FFTW_ESTIMATE);
    }

#ifdef _OPENMP
    #pragma omp parallel for
#endif
    for (index_t j = 0; j < n; j++) {
        const index_t T = get_thread_num();
        fftwf_plan planT = plan[T];
        sample_t* inT = bufin[T];
        sample_t* outT = bufout[T];
        gsl_vector_float_view Xj = gsl_vector_float_view_array(inT,m);
        gsl_vector_float_view TXj = gsl_vector_float_view_array(outT,m);
        gsl_matrix_float_get_row(&Xj.vector,X,j);
        dct2n_prenormalize(inT,w1,w2);
        fftwf_execute_r2r(planT,inT,outT);
        dct2n_postnormalize(outT,w1,w2);
        gsl_matrix_float_set_row(TX,j,&TXj.vector);
    }

    for (unsigned T = 0; T < NT; T++) {
        fftwf_destroy_plan(plan[T]);
        free(bufout[T]);
        free(bufin[T]);
    }
    fftwf_cleanup();
}

//---------------------------------------------------------------------------------------

static void idct2n_prenormalize(sample_t* inT, index_t w1, index_t w2) {
    const sample_t k00 = 2.0f;
    const sample_t k01 = (sample_t) M_SQRT2;
    const sample_t k10 = (sample_t) M_SQRT2;

    // upper-left element
    inT[0] *= k00;
    for (index_t j = 1; j < w2; ++j) // first row
        inT[j] *= k01;
    for (index_t i = 1; i < w1; ++i) // first column
        inT[w2*i] *= k10;
}

//---------------------------------------------------------------------------------------

static void idct2n_postnormalize(sample_t* outT, index_t w1, index_t w2) {
    const sample_t k = (float) 0.5/sqrt(w1*w2);
    const index_t m = w1*w2;
    for (index_t i = 0; i < m; ++i)
        outT[i] *= k;
}

//---------------------------------------------------------------------------------------

void idct2dn(gsl_matrix_float* X, const index_t w1, const index_t w2, const gsl_matrix_float* TX) {
    sample_t* bufin[MAX_THREADS]; // first, hopefully aligned!
    sample_t* bufout[MAX_THREADS]; // first, hopefully aligned!
    fftwf_plan plan[MAX_THREADS];
    const index_t m = TX->size2;
    const index_t n = TX->size1;
    assert(m == w1*w2);
    const index_t NT = get_num_threads();

    for (unsigned T = 0; T < NT; T++) {
        bufin[T] = (sample_t*) malloc(m*sizeof(sample_t));
        bufout[T] = (sample_t*) malloc(m*sizeof(sample_t));
        plan[T] = fftwf_plan_r2r_2d(w1,w2,bufin[T],bufout[T],FFTW_REDFT01,FFTW_REDFT01,FFTW_ESTIMATE);
    }

#ifdef _OPENMP
    #pragma omp parallel for
#endif
    for (index_t j = 0; j < n; j++) {
        const index_t T = get_thread_num();
        fftwf_plan planT = plan[T];
        sample_t* inT = bufin[T];
        sample_t* outT = bufout[T];
        gsl_vector_float_view TXj = gsl_vector_float_view_array(inT,m);
        gsl_vector_float_view Xj = gsl_vector_float_view_array(outT,m);
        gsl_matrix_float_get_row(&TXj.vector,TX,j);
        idct2n_prenormalize(inT,w1,w2);
        fftwf_execute_r2r(planT,inT,outT);
        idct2n_postnormalize(outT,w1,w2);
        gsl_matrix_float_set_row(X,j,&Xj.vector);
    }

    for (unsigned T = 0; T < NT; T++) {
        fftwf_destroy_plan(plan[T]);
        free(bufout[T]);
        free(bufin[T]);
    }
    fftwf_cleanup();
}

//===============================================================================

