/*
  * Copyright (c) 2019 Ignacio Francisco Ramírez Paulino and Ignacio Hounie
  * This program is free software: you can redistribute it and/or modify it
  * under the terms of the GNU Affero General Public License as published by
  * the Free Software Foundation, either version 3 of the License, or (at
  * your option) any later version.
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
  * General Public License for more details.
  * You should have received a copy of the GNU Affero General Public License
  *  along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
/**
 * \file paco_mapping.h
 * \brief Generic interface to signal<->patch mapping methods
 */
#ifndef PACO_MAPPING_H
#define PACO_MAPPING_H
#include "paco_types.h"


/**
 * \brief Function prototype for patch extraction operator. This corresponds to the OPERATOR $\mathcal{R}$ in the paper.
 * \param[out] X if *X is NULL, a new matrix is allocated and the patches are stored here
 * \param I input image
 * \param info parameters and configuration internal to the extraction operator (opaque)
 */
void extract_patches(gsl_matrix_float* X, const gsl_matrix_float* I);

/**
 * \brief Function prototype for patch stitching operator. This corresponds to the OPERATOR $\mathcal{S}$ in the paper.
 * \param[out] I stitched image. If *I is NULL, a new matrix is created
 * \param X input patches
 * \param info parameters and configuration internal to the stitching operator (opaque)
 */
void stitch_patches (gsl_matrix_float* I, const gsl_matrix_float* X);

/**
 * extracts a patch starting at (i,j)
 */
void extract_patch(sample_t* patch, const gsl_matrix_float* img, const index_t i, const index_t j, const index_t width, const index_t height);


/** @return true if all the samples in a patch are known, i.e., it has no missing samples.  */
char is_patch_complete(const gsl_matrix_float* mask, const index_t i0, const index_t j0, const index_t i1, const index_t j1);

/**
 * \bried Destroys the grid mapper.
 */
void paco_mapping_destroy();

index_t get_patch_width();

index_t get_patch_height();

index_t get_num_patch_samples();

index_t get_input_height();

index_t get_input_width();

index_t get_num_input_samples();

index_t get_num_mapped_samples();

index_t get_num_mapped_patches();

const index_t* get_idx_mapped_samples();

void extract_patches(gsl_matrix_float* P, const gsl_matrix_float* I);

void stitch_patches(gsl_matrix_float* I, const gsl_matrix_float* P);

#endif
