/*
  * Copyright (c) 2019 Ignacio Francisco Ramírez Paulino and Ignacio Hounie
  * This program is free software: you can redistribute it and/or modify it
  * under the terms of the GNU Affero General Public License as published by
  * the Free Software Foundation, either version 3 of the License, or (at
  * your option) any later version.
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
  * General Public License for more details.
  * You should have received a copy of the GNU Affero General Public License
  *  along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
/**
 * \file paco.h
 * \brief common interface to all PACO problems
 * The only function that needs to be declared differently for each implementation is the
 * problem initialization. 
 * Any new problem needs to declare the initializar function and define all the
 * ones declared here. The configuration is performed at linking time.
 */
#ifndef PACO_PROBLEM_H
#define PACO_PROBLEM_H
#include <gsl/gsl_matrix_float.h>
#include "paco_mapping.h"
#include "paco_iterate.h"
//
//----------------------------------------------------------------------------
// MAIN PROJECT PAGE
//----------------------------------------------------------------------------
//
/**
 * \mainpage
 *
 * \section summary Summary
 *
 * This is the C implementation of the PACO-DCT algorithm
 * described originally in [arXiv](https://arxiv.org/abs/1808.06942).
 *
 * \section architecture Overall architecture
 *
 * The code is designed so that different PACO-based problems
 * can be implemented by simply writing a few functions.
 *
 * The overall architecture is based on the 'Strategy' pattern,
 * that is, functions are passed as arguments to a generic implementation, paco_admm,
 * which is always the same. These functions are encapsulated along with
 * their related information into structures; such information is function dependent,
 * and thus they are referenced by generic (void*) poiners within the structure.
 *
 * For each particular problem, the user must provide a 'factory' function
 * which creates and populates the above structures with appropriate implementations
 * and function-specific structures, which are opaque to the user.
 *
 * \section more More information
 *
 * For more information see README.md in the root folder of the project.
 */
//
//----------------------------------------------------------------------------
// END MAIN PROJECT PAGE
//----------------------------------------------------------------------------
//

/**
 * change algorithm input signal
 */
void paco_problem_set_input(const gsl_matrix_float* input_samples);

/**
 * change algorithm reference signal
 */
void paco_problem_set_ref(const gsl_matrix_float* ref_samples);

/**
 * destroy all problem data 
 */
void paco_problem_destroy();

/**
 * Prototipe for _sentinel_ functions. These report intermediate results
 * during PACO iterations such as the current status of the optimizer,
 * partial outputs such as images or audio files, etc.
 */
void paco_sentinel(paco_iterate_st* iter);
/**
 * Prototype for initializing the ADMM iterate
 */
void paco_iter_init(paco_iterate_st* iter, const gsl_matrix_float* init_samples);

double paco_eval_f(const gsl_matrix_float* X);

void paco_prox_f( gsl_matrix_float *SA, const gsl_matrix_float* A, const double tau);

double paco_eval_g(const gsl_matrix_float* X);

void paco_prox_g( gsl_matrix_float *SA, const gsl_matrix_float* A, const double tau);




#endif
